/*
 *   catmbox: lock an mbox, and cat it to stdout, emptying it once
 *            a magic text appears on stdin.
 *   Copyright (C) 2007-2008 by Stefan Potyra <sistpoty@ubuntu.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sched.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>

#include "glue-io.h"

/* number of retries to acquire a lock on a mbox */
#define MAX_RETRIES 10
#define TIMEOUT_EMTPY 10
#define ACK_EMPTY "EMTPY\n"

static volatile bool terminate = false;
static volatile bool timeout = false;
static char *lock_link = NULL;
static char *mbox_lock = NULL;

static void
release_lock(const char *filename);

static void
sigterm(int signum)
{
	terminate = true;
	if (mbox_lock != NULL) {
		release_lock(mbox_lock);
	}
}

static void
sigalarm(int signum)
{
	timeout = true;
}

static char *
get_unique_name(const char *lockfile)
{
	static char name[4096];
	char *s;
	int len = sizeof(name);
	int ret;

	strncpy(name, lockfile, sizeof(name));
	name[sizeof(name) - 1] = '\0';

	ret = strlen(name);
	
	s = name + ret;
	len -= ret;

	if (len <= 0) {
		return NULL;
	}

	*s = '.';
	s++;
	len--;

	if (len <= 0) {
		return NULL;
	}

	ret = gethostname(s, len);
	if (ret < 0) {
		fprintf(stderr, "Cannot determine hostname\n");
		return NULL;
	}

	name[sizeof(name) - 1] = '\0';
	ret = strlen(s);
	len -= ret;
	s += ret;
	assert(len > 0);
	
	ret = snprintf(s, len, ".%d", getpid());
	assert((ret > 0) && (ret < len));
	len -= ret;
	
	if (len < 0) {
		return NULL;
	}

	return name;
}

static bool
create_link(const char *lockname)
{
	int fd;
	int ret;

	lock_link = get_unique_name(lockname);
	if (lock_link == NULL) {
		return false;
	}

	fd = open(lock_link, O_CREAT, S_IWUSR | S_IRUSR);
	if (fd < 0) {
		return false;
	}
	close(fd);
	
	ret = link(lock_link, lockname);
	if (ret < 0) {
		if (errno == EEXIST) {
			/* pretend success, we need to check again anyways. */
			return true;
		}

		fprintf(stderr, "Could not create lock %s: %s\n", lock_link,
				strerror(errno));
		return false;
	}

	return true;
}

static bool
acquire_lock(const char *lockname)
{
	int r;
	int retries;
	bool ret;
	struct stat lli;

	for (retries = 0; retries < MAX_RETRIES; retries++) {
		/* eventually create a unique file and a hard link to it as 
                   lockname */
		ret = create_link(lockname);
		if (! ret) {
			return false;
		}

		/* check if the lock_link actually is the lockname (i.e.
                   number of hard links==2, if not, retry */
		r = stat(lock_link, &lli);
		if (r < 0) {
			fprintf(stderr, "Cannot stat lock file %s: %s\n",
					lock_link, strerror(errno));
			return false;
		}

		if (lli.st_nlink < 2) {
			sleep(1);
			continue;
		}

		return true;
	}

	fprintf(stderr, "Could not create lock %s: Already locked an retry "
			"limit reached.\n", lockname);
	return false;
}

static void
release_lock(const char *filename)
{
	int ret;

	ret = unlink(filename);
	if (ret < 0) {
		fprintf(stderr, "Could not release lock %s: %s\n",
				filename, strerror(errno));
	}

	if (lock_link != NULL) {
		ret = unlink(lock_link);
	
		if (ret < 0) {
			fprintf(stderr, "Could not release lock %s: %s\n",
					lock_link, strerror(errno));
		}
	}
}


static bool
write_stdout(const char *buffer, ssize_t len)
{
	ssize_t ret;
	
	while(! terminate) {
		ret = write(1, buffer, len);
		if (ret < 0) {
			switch(errno) {
			case EAGAIN:
				sched_yield();
			case EINTR:
				break;
			default:
				return false;
			}
		}
		len -= ret;
		buffer += ret;

		if (len == 0) {
			return ! terminate;
		}
	}

	return ! terminate;
}

static bool
cat_mbox(const char *filename)
{
	int ret;

	FILE *f;
	static char line[1002];
	char *l;

	f = fopen(filename, "r");

	if (NULL == f) {
		fprintf(stderr, "Could not open mbox %s: %s\n", filename,
				strerror(errno));
		return false;
	}

	while ((! terminate) && (! feof(f))) {
		l = fgets(line, sizeof(line), f);
		if (l == NULL) {
			switch(errno) {
			case EAGAIN:
				sched_yield();
			case EINTR:
				continue;
			default:
				if (! feof(f)) {
					fprintf(stderr, "Could not read from"
						" mbox file %s: %s\n",
						filename, strerror(errno));
					fclose(f);
					return false;
				}
				break;
			}
			break;
		}
			
		/* escape lines starting with . */
		if (line[0] == '.') {
			write_stdout(".", 1);
		}
		write_stdout(line, strlen(line));
	}
	ret = fclose(f);
	assert(0 <= ret);

	/* end marker, line with a single "." */
	write_stdout(".\n", 2);
	return (! terminate);
}

static void
empty_mbox(const char *filename)
{
	int ret;

	if (terminate) {
		return;
	}

	ret = truncate(filename, 0);
	if (ret < 0) {
		fprintf(stderr, "Could not empty mbox %s: %s\n", filename, 
				strerror(errno));
	}
}

static void
usage(void)
{
	fprintf(stderr, "Syntax: catmbox [options]\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -m [M], --mbox=[M] \t specify mbox to read "
						"from\n");
	
	fprintf(stderr, "  -h, --help         \t display this text.\n");
	fprintf(stderr, "\n");
}

static void
register_sighandler(void)
{
	int ret;
	static struct sigaction sa_term;

	sa_term.sa_handler = sigterm;
	ret = sigfillset(&sa_term.sa_mask);
	assert(0 <= ret);
	ret = sigaction(SIGTERM, &sa_term, NULL);
	assert(0 <= ret);
	ret = sigaction(SIGINT, &sa_term, NULL);
	assert(0 <= ret);
	ret = sigaction(SIGPIPE, &sa_term, NULL);
	assert(0 <= ret);
}

static bool
wait_for_empty(const char *mbox)
{
	char buf[10];
	int r;
	bool ret;

	static struct sigaction sa_alarm;

	sa_alarm.sa_handler = sigalarm;
	/* allow SIGTERM to take priority */
	r = sigemptyset(&sa_alarm.sa_mask);
	assert(0 <= r);
	r = sigaction(SIGALRM, &sa_alarm, NULL);
	assert(0 <= r);

	alarm(TIMEOUT_EMTPY);

	assert(strlen(ACK_EMPTY) < sizeof(buf) - 1);

	ret = io_read(0, buf, strlen(ACK_EMPTY), &terminate);
	if (! ret) {
		fprintf(stderr, "reading ACK failed on stdin: %s\n",
				strerror(errno));
		return false;
	}

	alarm(0);

	buf[strlen(ACK_EMPTY)] = '\0';
	if (strcmp(buf, ACK_EMPTY) == 0) {
		empty_mbox(mbox);
	} 

	return true;
}

static bool
do_stuff(const char *mbox)
{
	ssize_t len;

	len = strlen(mbox) + strlen(".lock") + 1;
	mbox_lock = malloc(len);
	assert(mbox_lock);
	snprintf(mbox_lock, len, "%s.lock", mbox);


	if (! acquire_lock(mbox_lock)) {
		return false;
	}

	/* small race: locking succeeded and SIGTERM happens here */
	register_sighandler();

	if (! cat_mbox(mbox)) {
		if (! terminate) {
			release_lock(mbox_lock);
		}
		return false;
	}

	if (! wait_for_empty(mbox)) {
		if (! terminate) {
			release_lock(mbox_lock);
		}
		return false;
	}
	
	if (! terminate) {
		release_lock(mbox_lock);
	}

	return true;
}

int
main(int argc, char **argv) 
{
	bool ret;
	int r;
	char *mbox = NULL;

	static struct option option_long[] = {
		{ "help", 0, 0, 'h' },
		{ "mbox", 1, 0, 'm' },
		{ 0, 0, 0, 0 },
	};

	while (1) {
		r = getopt_long(argc, argv, "hm:", option_long, NULL);
		if (-1 == r) {
			break;
		}

		switch(r) {
		case 'm':
			mbox = optarg;
			break;
		case 'h': /* fall through */
		default:
			usage();
			return 1;
		}
	}

	if (mbox == NULL) {
		fprintf(stderr, "Must specify mbox file!\n");
		return 1;
	}

	ret = do_stuff(mbox);
	if (! ret) {
		if (lock_link != NULL) {
			unlink(lock_link);
		}	
		return 1;
	}

	return 0;
}
