/*
 *   getmailssh: get mail via ssh from an mbox file.
 *   Copyright (C) 2007  Stefan Potyra <sistpoty@ubuntu.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "glue-io.h"

/* FIXME */
#define HOST "faui31i.informatik.uni-erlangen.de"
#define USER "potyra"
#define REMOTE_MBOX "/home/inf3/potyra/mail/mbox_backup"
#define TARGET_MBOX "/home/stefan/target_mbox"
#define ACK_EMPTY "EMTPY\n"

/* determine number of elements in an array. */
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))

struct config {
	/* FIXME use ssh_command instead */
	const char *ssh_host;
	const char *ssh_user;
	const char *remote_mbox;
	const char *target_mbox;
	int ssh_fd;
};


static struct config cfg = {
	.ssh_host = HOST,
	.ssh_user = USER,
	.remote_mbox = REMOTE_MBOX,
	.target_mbox = TARGET_MBOX,
};

static volatile pid_t ssh_pid = 0;
static volatile bool terminate = false;

static void
sig_sigchild(int signum)
{
	pid_t pid;
	int status;

	if (ssh_pid > 0) {
		pid = waitpid(ssh_pid, &status, WNOHANG);

		if ((pid > 0) && WIFEXITED(status)) {
			if (WEXITSTATUS(status) != 0) {
				/* error occured */
				terminate = true;
			}
			
			ssh_pid = 0;
		}
	}
}

static void
launch_ssh(const struct config *cfg) __attribute__((__noreturn__));
static void
launch_ssh(const struct config *cfg)
{
	static char argv_buffer[8192];
	char *argv[20];
	char *walk = argv_buffer;
	int idx;
	
	snprintf(argv_buffer, sizeof(argv_buffer), 
		"ssh -C %s@%s %s",
		cfg->ssh_user, cfg->ssh_host, 
		"/home/inf3/potyra/bin/catmbox -m "
		"/home/inf3/potyra/mail/mbox_backup");

	argv_buffer[sizeof(argv_buffer) - 1] = '\0';

	for (idx = 0; idx < ARRAY_SIZE(argv); idx++) {
		argv[idx] = strsep(&walk, " ");

		if (argv[idx] == NULL) {
			break;
		}
	}

	if (idx == ARRAY_SIZE(argv)) {
		fprintf(stderr, "Too many arguments\n");
		exit(EXIT_FAILURE);
	}

	execvp(argv[0], argv);
	/* reached only on error */
	fprintf(stderr, "Could not execute: %s: %s\n", argv[0], 
		strerror(errno));
	exit(EXIT_FAILURE);
}

static bool
create_connection(struct config *cfg)
{
	/* iopipe[0]: child's end, [1]: parents end */
	int iopipe[2];
	int ret;

	ret = socketpair(PF_UNIX, SOCK_STREAM, 0, iopipe);
	if (ret < 0) {
		fprintf(stderr, "Could not create pipe: %s\n", 
			strerror(errno));

		return false;
	}

	ssh_pid = fork();
	switch(ssh_pid) {
	case -1:
		fprintf(stderr, "Could not fork: %s\n",  strerror(errno));
		ret = close(iopipe[0]);
		assert(0 <= ret);
		ret = close(iopipe[1]);
		assert(0 <= ret);
		return false;
	case 0: 
		/* child process */
		ret = close(0);
		assert(0 <= ret);
		ret = dup2(iopipe[0], 0);
		if (ret < 0) {
			fprintf(stderr, "dup2 failed: %s\n",
				strerror(errno));
			exit(EXIT_FAILURE);
		}

		ret = close(1);
		assert(0 <= ret);
		ret = dup2(iopipe[0], 1);
		if (ret < 0) {
			fprintf(stderr, "dup2 failed: %s\n",
				strerror(errno));
			exit(EXIT_FAILURE);
		}
		ret = close(iopipe[0]);
		assert(0 <= ret);
		ret = close(iopipe[1]);
		assert(0 <= ret);

		launch_ssh(cfg);
		/* not reached */
	
		break;
	default:
		/* parent process */
		ret = close(iopipe[0]);
		assert(0 <= ret);
		cfg->ssh_fd = iopipe[1];
		break;
	}

	return true;
}


static void
destroy_connection(const struct config *cfg)
{
	pid_t pid;
	int status;
	int ret;

	if (ssh_pid > 0) {
		do {
			pid = waitpid(ssh_pid, &status, 0);
			if (pid == -1) {
				fprintf(stderr, "waitpid: %s\n", 
					strerror(errno));
				break;
			}
		} while (! WIFEXITED(status));
	}

	if (cfg->ssh_fd > 0) {
		ret = close(cfg->ssh_fd);
		if (ret < 0) {
			fprintf(stderr, "close ssh_fd failed: %s\n", 
					strerror(errno));
		}
	}
}

static bool
forward_mbox(const struct config *cfg, int target_fd)
{
	char line[4010];
	char *l;
	ssize_t len;
	bool ret;
	
	while (! terminate) {
		l = line;
		ret = io_readline(cfg->ssh_fd, line, sizeof(line), 
				&terminate);
		if (! ret) {
			if (terminate) {
				return false;
			}

			fprintf(stderr, "Reading from ssh failed: %s\n",
					strerror(errno));
			return false;
		}

		if (strncmp(".\n", line, strlen(".\n")) == 0) {
			/* end token found. */
			return true;
		}

		if (l[0] == '.') {
			l++;
		}
		
		len = strlen(l);
		ret = io_write(target_fd, l, strlen(l), &terminate);
		if (! ret) {
			fprintf(stderr, "Write failed: %s\n",
				strerror(errno));
			break;
		}

	}

	/* only reached on error */
	return false;
}

static void
send_empty_cmd(int write_fd)
{
	int ret;

	ret = write(write_fd, &ACK_EMPTY, strlen(ACK_EMPTY));
	if (ret != strlen(ACK_EMPTY)) {
		fprintf(stderr, "Writing acknowledge to empty source mbox"
				" failed: %s\n", strerror(errno));
	}
}

static int
open_target_mbox(const struct config *cfg)
{
	int target_fd;
	/* TODO use locking? */

	target_fd = open(cfg->target_mbox, O_CREAT | O_EXCL | O_WRONLY, 
			S_IRUSR | S_IWUSR);
	if (target_fd < 0) {
		fprintf(stderr, "Cannot open target mbox %s: %s\n",
			cfg->target_mbox, strerror(errno));
		return -1;
	}

	return target_fd;
}

static void
setup_sighnd(void)
{
	int ret;
	static struct sigaction sa_child;

	sa_child.sa_handler = sig_sigchild;
	ret = sigfillset(&sa_child.sa_mask);
	assert(0 <= ret);
	ret = sigaction(SIGCHLD, &sa_child, NULL);
	assert(0 <= ret);
}

int
main(int argc, char **argv)
{
	int target_fd;
	int ret;

	target_fd = open_target_mbox(&cfg);
	if (target_fd < 0) {
		return 1;
	}

	setup_sighnd();

	if (! create_connection(&cfg)) {
		close(target_fd);
		return 1;
	}

	if (! forward_mbox(&cfg, target_fd)) {
		destroy_connection(&cfg);
		close(target_fd);
		return 1;
	}

	ret = close(target_fd);
	assert(0 <= ret);

	send_empty_cmd(cfg.ssh_fd);
	destroy_connection(&cfg);

	return 0;
}
