/*
 *   getmailssng, rewrite of getmailssh. (FIXME title)
 *   Copyright (C) 2008 by Stefan Potyra <sistpoty@ubuntu.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

/** class for a generic stream source */
struct stream_src {
	/** method to open a stream source. 
         *  @param s class instance.
         *  @param arg optional string argument 
         *  @return true on success, false otherwise 
         */
	bool (*open)(struct stream_src *s, const char *arg);

	/** method to close a stream source. 
         *  @param s class instance. */
	void (*close)(struct stream_src *s);

	/** read a line from the stream source.
         *  @param s class instance.
         *  @param len read maximum len bytes.
         *  @param dest store read line in dest.
         *  @return -1 on error, 0 for eof and 1 for success.
         */
	int (*readln)(struct stream_src *s, size_t len, char *dest);

	/** empty the stream source.
         *  @param s class instance.
         */
	void (*empty)(struct stream_src *s);

	/** private class data */
	void *_s;
};

/* --------- file stream handling ----------- */

/** local class data */
struct file_stream_src {
	/** source of the stream if opened. */
	FILE *src;
	/** path to file */
	const char *path;
};

static bool
file_stream_src_open(struct stream_src *s, const char *arg)
__attribute__((__used__));
static bool
file_stream_src_open(struct stream_src *s, const char *arg)
{
	struct file_stream_src *fs = (struct file_stream_src*)s->_s;
	assert(arg != NULL);
	
	fs->path = arg;
	fs->src = fopen(arg, "r");
	if (fs->src == NULL) {
		fprintf(stderr, "Failed to open file %s: %s\n", arg, 
			strerror(errno));
		return false;
	}

	return true;
}

static void
file_stream_src_close(struct stream_src *s)
__attribute__((__used__));
static void
file_stream_src_close(struct stream_src *s)
{
	int ret;
	struct file_stream_src *fs = (struct file_stream_src*)s->_s;

	assert(fs->src != NULL);
	assert(fs->path != NULL);

	ret = fclose(fs->src);
	fs->src = NULL;
	
	if (ret != 0) {
		fprintf(stderr, "Failed to close file %s: %s\n",
			fs->path,
			strerror(errno));
	}
}

static int
file_stream_src_readln(struct stream_src *s, size_t len, char *dest)
__attribute__((__used__));
static int
file_stream_src_readln(struct stream_src *s, size_t len, char *dest)
{
	struct file_stream_src *fs = (struct file_stream_src*)s->_s;
	char *ret;
	int eb;

	assert(fs->src != NULL);
	assert(fs->path != NULL);

	errno = 0;
	ret = fgets(dest, len, fs->src);
	if (ret == NULL) {
		eb = errno;
		if (feof(fs->src)) {
			return 0;
		}

		fprintf(stderr, "Error while reading from %s: %s\n",
			fs->path, strerror(eb));
		return -1;
	}
	return 1;
}

static void
file_stream_src_empty(struct stream_src *s)
__attribute__((__used__));
static void
file_stream_src_empty(struct stream_src *s)
{
	struct file_stream_src *fs = (struct file_stream_src*)s->_s;
	int ret;

	/* must not be opened! */
	assert(fs->src == NULL);
	assert(fs->path != NULL);

	/* or remove the file instead? */
	ret = truncate(fs->path, 0);
	if (ret != 0) {
		fprintf(stderr, "Cannot truncate file %s: %s\n",
			fs->path, strerror(errno));
	}
}



int
main(int argc, char **argv)
{
	return 0;
}
