/*
 *   glue-io.c: basic io operations on file descriptors.
 *   Copyright (C) 2007  Stefan Potyra <sistpoty@ubuntu.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <stdbool.h>
#include <unistd.h>
#include <errno.h>
#include <sched.h>
#include <assert.h>
#include <stdio.h>

bool
io_read(
	int fd, 
	char *buffer, 
	ssize_t len, 
	volatile bool *terminate
)
{
	ssize_t ret;

	while (! (*terminate)) {
		ret = read(fd, buffer, len);
		if (ret == 0) {
			/* eof/broken pipe */
			return false;
		}
		if (ret < 0) {
			switch(errno) {
			case EAGAIN:
				sched_yield();
			case EINTR:
				break;
			default:
				return false;
			}
		} else {
			len -= ret;
			buffer += ret;
		}

		if (len == 0) {
			return true;
		}
		
	}
	return ! (*terminate);
}

bool
io_write(
	int fd, 
	const char *buffer,
	ssize_t len, 
	volatile bool *terminate
)
{
	ssize_t ret;
	
	while(! (*terminate)) {
		ret = write(fd, buffer, len);
		if (ret < 0) {
			switch(errno) {
			case EAGAIN:
				sched_yield();
			case EINTR:
				break;
			default:
				return false;
			}
		} else {
			len -= ret;
			buffer += ret;
		}

		if (len == 0) {
			return true;
		}
	}

	return ! (*terminate);
}

bool
io_readline(
	int fd,
	char *buffer,
	ssize_t len,
	volatile bool *terminate
)
{
	bool ret;

	assert(1 < len); /* cannot store nl + \0 with len < 2 */

	while (! (*terminate)) {
		ret = io_read(fd, buffer, 1, terminate);
		if (! ret) {
			return false;
		}

		len--;
		if (len == 0) {	
			fprintf(stderr, "buffer too small for line\n");
			return false;
		}

		if (buffer[0] == '\n') {
			buffer[1] = '\0';
			return true;
		}

		buffer++;
	}
	
	/* *terminate == true */
	return false;
}
