/*
 *   mbox2rsmtp: convert an mbox file to rsmtp commands.
 *   Copyright (C) 2007-2008 by Stefan Potyra <sistpoty@ubuntu.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sched.h>

struct config {
	const char *receiver;
	const char *localdomain;
	const char *smtpcommand;
	const char *mbox_file;
	unsigned int max_transactions;
};

static int child_fd;
static int child_pid;

static bool
launch_smtp(const struct config *cfg)
{
	int iopipe[2];
	int ret;

	ret = socketpair(PF_UNIX, SOCK_STREAM, 0, iopipe);
	if (ret < 0) {
		fprintf(stderr, "Could not create pipe: %s\n", 
			strerror(errno));

		return false;
	}


	child_pid = fork();
	switch(child_pid) {
	case -1:
		fprintf(stderr, "Could not fork: %s\n",  strerror(errno));
		close(iopipe[0]);
		close(iopipe[1]);
		return false;
	case 0: 
		/* child process */
		ret = dup2(iopipe[0], 0);
		if (ret < 0) {
			fprintf(stderr, "dup2 failed: %s\n",
				strerror(errno));
			exit(EXIT_FAILURE);
		}
		ret = close(1);
		assert(0 <= ret);

		open("/dev/null", O_WRONLY);
		if (ret < 0) {
			fprintf(stderr, "Could not open /dev/null: %s\n",
				strerror(errno));
			exit(EXIT_FAILURE);
		}
		ret = close(iopipe[1]);
		assert(0 <= ret);

		ret = execlp(cfg->smtpcommand, cfg->smtpcommand, NULL);

		/* not reached except error */
		if (ret < 0) {
			fprintf(stderr, "Could not execute smtp command "
					" %s: %s\n", 
				cfg->smtpcommand, 
				strerror(errno));

			exit(EXIT_FAILURE);
		}

		break;
	default:
		/* parent process */
		close(iopipe[0]);
		child_fd = iopipe[1];
		break;
	}

	return true;
}

static bool
write_child(const char *msg)
{	
	size_t len;
	int ret;

	len = strlen(msg);
	for (;;) {
		ret = write(child_fd, msg, len);
		if (ret == len) {
			break;
		}

		if (ret == 0) {
			return false;
		}

		if ((ret < 0) && (errno == EINTR)) {
			continue;
		}

		if ((ret < 0) && (errno == EAGAIN)) {
			sched_yield();
		}
		
		if (ret > 0) {
			len -= ret;
			msg += ret;
		}
	}

	return true;
}

static void
kill_smtp(void)
{
	int ret;
	int status;
	pid_t pid;

	ret = close(child_fd);
	assert(0 <= ret);
	
	do {
		pid = waitpid(child_pid, &status, 0);
	} while (! WIFEXITED(status));
}

static char*
get_email_address(char *line)
{
	int len = strlen(line);
	int pos;

	for (pos = 0; pos <= len; pos++) {
		if (line[pos] == ' ') {
			break;
		}
	}

	assert(pos > 0);
	line[pos] = '\0';
	return line;
}

static char*
get_from_address(char *line)
{
	int res;

	res = strncmp("From ", line, strlen("From "));
	if (res != 0) {
		return NULL;
	}

	return get_email_address(line + strlen("From "));
}

static bool
print_helo(const struct config *cfg)
{
	bool ret = true;

	ret &= write_child("HELO ");
	ret &= write_child(cfg->localdomain);
	ret &= write_child("\n");

	return ret;
}

static bool
print_header(const struct config *cfg, const char *from)
{
	bool ret = true;

	ret &= write_child("MAIL FROM: ");
	ret &= write_child(from);
	ret &= write_child("\n");

	ret &= write_child("RCPT TO: ");
	ret &= write_child(cfg->receiver);
	ret &= write_child("\n");
	ret &= write_child("DATA\n");

	return ret;
}

static bool
print_footer(void)
{
	bool ret = true;

	ret &= write_child(".\n");
	return ret;
}

static bool
print_quit(void)
{
	bool ret = true;

	ret &= write_child("QUIT\n");

	return ret;
}

static void
report_error(const char *mbox)
{
	fprintf(stderr, "Could not open mbox file %s: %s\n", mbox, 
		strerror(errno));
}

static int
parse_mbox(const struct config *cfg)
{
	FILE *f;
	static char line[1002];
	char *from;
	int msg_counter = 0;
	bool ret = true;
	char *l;

	f = fopen(cfg->mbox_file, "r");
	if (NULL == f) {
		report_error(cfg->mbox_file);
		return 1;
	}

	while (! feof(f) && ret) {
		l = fgets(line, sizeof(line), f);
		if (l == NULL) {
			break;
		}

		from = get_from_address(l);
		if (from != NULL) {
			if (msg_counter > 0) {
				ret &= print_footer();
			}

			if (msg_counter == cfg->max_transactions) {
				ret &= print_quit();
				msg_counter = 0;

				kill_smtp();
				/* TODO check for errors */
				launch_smtp(cfg);
			}
			
			if (msg_counter == 0) {
				ret &= print_helo(cfg);
			}

			msg_counter++;
			ret &= print_header(cfg, from);
			continue;
		}
		
		if ('.' == line[0]) {
			ret &= write_child(".");
			ret &= write_child(l);
			continue;
		}
		ret &= write_child(line);
	}

	if ((msg_counter) > 0 && ret) {
		ret &= print_footer();
		ret &= print_quit();
	}

	fclose(f);

	if (! ret) {
		return -1;
	}
	return 0;
}

static bool
determine_user(struct config *cfg)
{
	struct passwd *p;
	static char rec_addr[200];
	uid_t u;

	u = getuid();
	p = getpwuid(u);
	if (p == NULL) {
		fprintf(stderr, "Could not determine username. Please"
				" specify receiver address.\n");
		return false;
	}
	
	snprintf(rec_addr, sizeof(rec_addr), "%s@localhost", p->pw_name);
	rec_addr[sizeof(rec_addr) - 1] = '\0';
	cfg->receiver = rec_addr;

	return true;
}

static void
usage(void)
{
	fprintf(stderr, "Syntax: mbox2rsmtp [options]\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -l [L], --localdomain=[L]\t set localdomain for\n");
	fprintf(stderr, "                           \t rsmtp HELO command.\n");
	fprintf(stderr, "  -r [R], --receiver=[R]   \t set recipient of\n");
	fprintf(stderr, "                           \t mail for rsmtp.\n");
	fprintf(stderr, "  -s [S], --smtp=[S]       \t set rsmtp program\n");
	fprintf(stderr, "                           \t default=rsmtp\n");
	fprintf(stderr, "  -m [M], --mbox=[M]	    \t specify mbox to\n");
	fprintf(stderr, "                           \t read from\n");
	fprintf(stderr, "  -b [B], --batch-limit=[B]\t maximum number of\n");
	fprintf(stderr, "                           \t smtp transactions\n");
	fprintf(stderr, "  -h, --help               \t display this text.\n");
	fprintf(stderr, "\n");
	return;
}

int
main(int argc, char **argv) 
{
	int ret;
	int r;

	static struct option option_long[] = {
		{ "receiver", 1, 0, 'r' },
		{ "localdomain", 1, 0, 'l' },
		{ "help", 0, 0, 'h' },
		{ "smtp", 1, 0, 's' },
		{ "mbox", 1, 0, 'm' },
		{ "batch-limit", 1, 0, 'b' },
		{ 0, 0, 0, 0 },
	};

	static struct config cfg = {
		.receiver = NULL,
		.localdomain = "mbox2rsmtp.localhost",
		.smtpcommand = "/usr/sbin/rsmtp",
		.mbox_file = NULL,
		.max_transactions = 10,
	};

	while (1) {
		r = getopt_long(argc, argv, "hr:l:s:m:b:", option_long, NULL);
		if (-1 == r) {
			break;
		}

		switch(r) {
		case 'r':
			cfg.receiver = optarg;
			break;
		case 'l':
			cfg.localdomain = optarg;
			break;
		case 's':
			cfg.smtpcommand = optarg;
			break;
		case 'm':
			cfg.mbox_file = optarg;
			break;
		case 'b':
			cfg.max_transactions = atoi(optarg);
			break;
		case 'h': /* fall through */
		default:
			usage();
			return 1;
		}
	}

	if (cfg.receiver == NULL) {
		if (! determine_user(&cfg)) {
			return 1;
		}
	}

	if (cfg.mbox_file == NULL) {
		fprintf(stderr, "Must specify mbox file!\n");
		return 1;
	}

	if (! launch_smtp(&cfg)) {
		return 1;
	}

	ret = parse_mbox(&cfg);

	kill_smtp();
	return ret;
}
